"""
Copyright (c) 2023-2025 Biomedical Primate Research Centre, the Netherlands.
All rights reserved.
"""

import sys
import traceback
from .logger import console_logger, file_logger

console_log = console_logger(__name__)
file_log = file_logger(__name__)


def log_error():
    """
    Decorator to log errors that occur in the decorated function.

    Returns:
        function: A decorator function that wraps the original function.
    """
    def decorator(func):
        """
        Inner decorator function.

        Args:
            func (function): The function to be decorated.

        Returns:
            function: The wrapped function with error logging.
        """
        def wrapper(*args, **kwargs):
            """
            Wrapper function that executes the original function and logs any exceptions.

            Args:
                *args: Variable length argument list for the original function.
                **kwargs: Arbitrary keyword arguments for the original function.

            Returns:
                Any: The return value of the original function, if no exception occurs.

            Raises:
                SystemExit: Exits the program with status 1 if an exception occurs.
            """
            try:
                return func(*args, **kwargs)
            except Exception as e:
                # Haal de laatste foutinformatie op
                tb = traceback.extract_tb(e.__traceback__)
                last_call = tb[-1]  # Pak het laatste traceback-object
                file_name = last_call.filename
                line_number = last_call.lineno

                # Bouw het foutbericht op
                error_message = (
                    f"Error in '{func.__name__}' in file '{file_name}', line {line_number}: {e}"
                )

                # Log foutberichten naar de console en het bestand
                console_log.error(error_message)
                file_log.error(error_message)

                if hasattr(e, 'value'):
                    context_message = f"Additional info: {e.value}"
                    console_log.error(context_message)
                    file_log.error(context_message)

                sys.exit(1)
        return wrapper
    return decorator



