"""
Copyright (c) 2023-2025 Biomedical Primate Research Centre, the Netherlands.
All rights reserved.
"""
from typing import Union

from pathlib import Path
import pandas as pd
import yaml
import zipfile
import argparse
import psutil

from .logger import console_logger, file_logger
from .property import log_error

console_log = console_logger(__name__)
file_log = file_logger(__name__)

def make_dir(path: Union[str, Path]) -> Path:
    """
    Ensures the specified directory exists by creating it if necessary. Checks if the directory exists,
    and if it doesn't, creates the directory along with any necessary parent directories. Logs the creation
    or existence of the directory. If an error occurs during the directory creation, raises an `OSError`.

    Args:
        path (str): The path of the directory to be created.

    Returns:
        Path: The path of the created or existing directory.

    Raises:
        OSError: If the directory cannot be created due to a system-related error.
    """
    try:
        if Path(path).exists():
            file_log.info(f"Directory already exists: {path}")
        else:
            file_log.info(f"Directory created: {path}")
        Path(path).mkdir(parents=True, exist_ok=True)
    except Exception as e:
        file_log.error(f"Failed to create directory {path}: {e}")
        file_log.error(f"Failed to create directory {path}: {e}")
        raise OSError(f"Failed to create directory {path}") from e
    return Path(path)


def validate_directory(path: Union[str, Path]) -> Union[str, Path]:
    """
    Validates that the specified directory exists. Checks if the provided path corresponds to an existing directory.
    If the directory does not exist, raises an `argparse.ArgumentTypeError`.

    Args:
        path (str): Path to the directory to validate.

    Returns:
        str: The validated directory path.

    Raises:
        argparse.ArgumentTypeError: If the directory does not exist.
    """
    if not Path(path).is_dir():
        raise argparse.ArgumentTypeError(f"The directory {path} does not exist. Try another directory!")
    return path


def validate_file(path: str) -> str:
    """
    Validates that the specified file exists. Checks if the provided path corresponds to an existing file.
    If the file does not exist, raises an `argparse.ArgumentTypeError`.

    Args:
        path (str): Path to the file to validate.

    Returns:
        str: The validated file path.

    Raises:
        argparse.ArgumentTypeError: If the file does not exist.
    """
    if not Path(path).is_file():
        raise argparse.ArgumentTypeError(f"The file {path} does not exist. Try another file, please!")
    return path


def validate_input(path: str) -> str:
    """
    Validates the input directory, ensuring it exists and contains FASTA files. Checks that the specified directory
    exists and contains at least one FASTA file (with extensions .fasta, .fa, or .fna). If the directory is empty
    or contains no FASTA files, raises an `argparse.ArgumentTypeError`.

    Args:
        path (str): Path to the input directory.

    Returns:
        str: The validated input directory path.

    Raises:
        argparse.ArgumentTypeError: If the directory is empty or contains no FASTA files.
    """

    input_path = Path(path).resolve()
    validate_directory(str(input_path))
    if not any(entry.is_file() for ext in ["*.fasta", "*.fa", "*.fna"] for entry in input_path.glob(ext)):
        raise argparse.ArgumentTypeError(
            f"""The directory {
                input_path} is empty or does not contain any FASTA files!"""
        )
    return str(input_path)


@log_error()
def validate_metadata_coverage(input_dir: str, metadata_dir: str) -> bool:
    input_files = []
    for pattern in ["*.fasta", "*.fa", "*.fna"]:
        input_files.extend([f.name for f in Path(input_dir).glob(pattern)])

    metadata_df = pd.read_excel(metadata_dir)
    expected_files = set(metadata_df['Accession'])

    missing_files = []
    for input_file in input_files:
        if not any(basename in input_file for basename in expected_files):
            missing_files.append(input_file)

    if missing_files:
        console_log.warning("Warning: The following input files are not covered in the metadata:")
        for f in missing_files:
            console_log.warning(f"{input_dir}/{f}")
        return False
    return True


def load_config(path: Union[str, Path]) -> dict:
    """
    Loads a configuration file (config.yaml) located in the 'config' directory of the current working directory.

    Args:
        path (str): The current working directory.

    Returns:
        dict: Dictionary containing the configuration settings, including chromosomes of interest and their respective flanking genes.

    Raises:
        Exception: If the configuration file cannot be loaded, logs the error and raises an exception.
    """
    try:
        with open(path) as f:
            config = yaml.safe_load(f)
            file_log.info("Config file loaded successfully")
            return config
    except Exception as e:
        file_log.error(f"Failed to load config file: {e}")
        raise


def unzip_file(file_path: Union[str, Path], unzip_path: Union[str, Path]) -> None:
    """
    Extracts a zip file to the specified directory.

    Args:
        file_path (str or Path): Path to the zip file to be extracted.
        unzip_path (str or Path): Directory where the zip file's contents will be extracted.

    Raises:
        Exception: If the extraction fails, logs the error and raises an exception.
    """
    extract_to_path = Path(unzip_path)
    try:
        with zipfile.ZipFile(file_path, 'r') as zip_ref:
            zip_ref.extractall(extract_to_path)
        file_log.info(f'Extracted {file_path} to {extract_to_path}')
        file_log.info(f'Extracted {file_path} to {extract_to_path}')
    except Exception as e:
        file_log.error(f"Failed to extract {file_path} to {extract_to_path}: {e}")
        file_log.error(f"Failed to extract {file_path} to {extract_to_path}: {e}")
        raise


def calculate_available_resources(max_cores: int, threads: int, memory_per_process: int = 12, buffer_percentage: int = 10) -> int:
    """
    Calculate the maximum number of parallel jobs that can be run based on the available system resources.

    This function determines the available physical memory (RAM) after applying a buffer for system overhead
    and calculates the maximum number of processes that can be run concurrently, considering both memory
    and CPU core limitations.

    Args:
        max_cores (int): Total number of CPU cores available.
        cores_per_process (int): Number of CPU cores required per process.
        memory_per_process (int, optional): Memory (in GB) required per process. Defaults to 16 GB.
        buffer_percentage (int, optional): Percentage of total memory reserved for system and overhead. Defaults to 10%.

    Returns:
        int: Maximum number of processes that can be run in parallel based on available memory and cores.
    """
    memory = psutil.virtual_memory().available
    total_memory = psutil.virtual_memory().total

    buffer_memory = total_memory * (buffer_percentage / 100)
    available_memory_after_buffer = (memory - buffer_memory) / (1024 ** 3)
    available_memory = max(available_memory_after_buffer, 0)

    max_jobs_by_memory = max(1, int(available_memory // memory_per_process))
    max_jobs_by_cores = max_cores // threads

    file_log.info(f"Running with max workers: {min(max_jobs_by_memory, max_jobs_by_cores)}")
    file_log.info(f"Running with max workers: {min(max_jobs_by_memory, max_jobs_by_cores)}")
    return min(max_jobs_by_memory, max_jobs_by_cores)
